# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

# Writing an SQW file is fairly slow.
# So the tests in this module use module-scoped fixtures to write a single file
# that is shared between all tests.
# Function-scoped fixtures allow accessing that file for reading.

import itertools
from collections.abc import Generator
from pathlib import Path

import numpy as np
import numpy.typing as npt
import pytest
import sciline
import scipp as sc
import scipp.testing
import scippnexus as snx
from scippneutron.io import sqw

from ess import bifrost
from ess.bifrost.data import (
    lookup_table_simulation,
    simulated_elastic_incoherent_with_phonon,
)
from ess.spectroscopy.types import (
    EnergyBins,
    Filename,
    LookupTableFilename,
    LookupTableRelativeErrorThreshold,
    NeXusDetectorName,
    OutFilename,
    PreopenNeXusFile,
    SampleRun,
    SQWBinSizes,
    UncertaintyBroadcastMode,
)

N_DETECTORS = 3
N_ANGLES = 180

BIN_SIZES = {'u1': 6, 'u2': 7, 'u3': 8, 'u4': 9}
ENERGY_BIN_SIZE = 13

# Q projections
U = sc.vector([1, 0, 0], unit="1/angstrom")
V = sc.vector([0, 1, 0], unit="1/angstrom")
W = sc.vector([0, 0, 1], unit="1/angstrom")

N_PIXELS_PER_DETECTOR = 300  # Fixed, not a parameter!


@pytest.fixture(scope='module')
def detector_names() -> list[NeXusDetectorName]:
    with snx.File(simulated_elastic_incoherent_with_phonon()) as f:
        detector_names = list(f['entry/instrument'][snx.NXdetector])
    return detector_names[:N_DETECTORS]


@pytest.fixture(scope='module')
def sample() -> sqw.SqwIXSample:
    return sqw.SqwIXSample(
        name="Vibranium",
        lattice_spacing=sc.vector([2.86, 2.86, 2.86], unit="angstrom"),
        lattice_angle=sc.vector([90.0, 90.0, 90.0], unit="deg"),
    )


@pytest.fixture(scope='module')
def common_workflow(
    detector_names: list[NeXusDetectorName], sample: sqw.SqwIXSample
) -> sciline.Pipeline:
    wf = bifrost.BifrostSimulationWorkflow(detector_names)

    wf[Filename[SampleRun]] = simulated_elastic_incoherent_with_phonon()
    wf[LookupTableFilename] = lookup_table_simulation()
    wf[LookupTableRelativeErrorThreshold] = {
        'detector': np.inf,
        '110_frame_3': np.inf,
    }
    wf[PreopenNeXusFile] = PreopenNeXusFile(True)
    wf[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.drop
    wf[sqw.SqwIXSample] = sample
    wf[EnergyBins] = ENERGY_BIN_SIZE
    wf[SQWBinSizes] = BIN_SIZES

    return wf


@pytest.fixture(scope='module')
def write_file(
    common_workflow: sciline.Pipeline, tmp_path_factory: pytest.TempPathFactory
) -> Path:
    path = tmp_path_factory.mktemp("bifrost-sqw").joinpath("bifrost-simulated.sqw")
    workflow = common_workflow.copy()
    workflow[OutFilename] = path
    workflow.bind_and_call(bifrost.io.sqw.save_sqw)
    return path


@pytest.fixture
def output_file(write_file: Path) -> Generator[sqw.Sqw, None, None]:
    with sqw.Sqw.open(write_file) as file:
        yield file


def test_save_sqw_writes_instrument_metadata(output_file: sqw.Sqw) -> None:
    instruments = output_file.read_data_block("experiment_info", "instruments")

    assert len(instruments) == N_ANGLES
    # All instruments are the same:
    for instrument in instruments[1:]:
        sc.testing.assert_identical(instrument, instruments[0])

    instrument = instruments[0]
    assert instrument.name == "BIFROST"
    sc.testing.assert_identical(instrument.source.frequency, sc.scalar(14.0, unit="Hz"))


def test_save_sqw_writes_sample_metadata(
    output_file: sqw.Sqw, sample: sqw.SqwIXSample
) -> None:
    samples = output_file.read_data_block("experiment_info", "samples")

    assert len(samples) == N_ANGLES
    # All samples are the same:
    for s in samples:
        sc.testing.assert_identical(s, sample)


def test_save_sqw_writes_experiment_metadata(output_file: sqw.Sqw) -> None:
    experiments = output_file.read_data_block("experiment_info", "expdata")

    assert len(experiments) == N_ANGLES
    # N unique run ids
    assert len({experiment.run_id for experiment in experiments}) == N_ANGLES
    for experiment in experiments:
        assert experiment.emode == sqw.EnergyMode.indirect
        assert experiment.u == U
        assert experiment.v == V


def test_save_sqw_writes_dnd_metadata(
    output_file: sqw.Sqw, sample: sqw.SqwIXSample
) -> None:
    metadata = output_file.read_data_block("data", "metadata")

    np.testing.assert_array_equal(
        metadata.axes.n_bins_all_dims.values, list(BIN_SIZES.values())
    )

    sc.testing.assert_identical(metadata.proj.u, U)
    sc.testing.assert_identical(metadata.proj.v, V)
    sc.testing.assert_identical(metadata.proj.w, W)
    assert metadata.proj.type == "aaa"
    assert not metadata.proj.non_orthogonal

    sc.testing.assert_identical(metadata.proj.lattice_spacing, sample.lattice_spacing)
    sc.testing.assert_identical(metadata.proj.lattice_angle, sample.lattice_angle)


def test_save_sqw_writes_dnd_data(output_file: sqw.Sqw) -> None:
    dnd = output_file.read_data_block("data", "nd_data")
    values, errors, counts = dnd
    # Fortran order
    expected_shape = tuple(BIN_SIZES.values())[::-1]
    assert values.shape == expected_shape
    assert errors.shape == expected_shape
    assert counts.shape == expected_shape


def test_save_sqw_writes_pixel_data(output_file: sqw.Sqw) -> None:
    metadata = output_file.read_data_block("data", "metadata")
    dnd = output_file.read_data_block("data", "nd_data")
    pix = output_file.read_data_block("pix", "data_wrap")

    assert pix.shape == (
        N_DETECTORS * N_PIXELS_PER_DETECTOR * N_ANGLES * ENERGY_BIN_SIZE,
        9,
    )

    n_pix = dnd[2].astype(int)
    check_pixels_in_bin_ranges(pix, metadata, n_pix)
    check_pixel_indices_in_ranges(pix)


def check_pixels_in_bin_ranges(
    pix: npt.NDArray[np.float32], metadata: sqw.SqwDndMetadata, n_pix: npt.NDArray[int]
) -> None:
    """Check that all pixels are within the bin edges defined in the dnd metadata."""
    img_range = metadata.axes.img_range
    n_bins = metadata.axes.n_bins_all_dims.values
    u_edges = [
        sc.linspace(f'u{i}', img_range[i][0], img_range[i][1], n_bins[i] + 1)
        for i in range(len(n_bins))
    ]

    bin_offsets = np.r_[0, np.cumsum(n_pix.flat)]
    for bin_index, (l, k, j, i) in enumerate(  # noqa: E741
        itertools.product(*(range(nb) for nb in n_bins[::-1]))
    ):
        pix_slice = slice(bin_offsets[bin_index], bin_offsets[bin_index + 1])
        if pix_slice.stop <= pix_slice.start:
            continue  # empty bin => no pixels to check
        for col, index in enumerate((i, j, k, l)):
            p = pix[pix_slice, col]
            p_min = p.min()  # actual range
            p_max = p.max()
            e_min = u_edges[col][index].value  # expected range
            e_max = u_edges[col][index + 1].value
            assert p_min >= e_min - 1e-6  # small offset to account for f32 rounding
            assert p_max < e_max + 1e-6


def check_pixel_indices_in_ranges(pix: npt.NDArray[np.float32]) -> None:
    """Check that all indices in the pixel data are in the correct ranges."""

    irun = pix[:, 4]
    idet = pix[:, 5]
    ien = pix[:, 6]

    # 1-based indices!
    assert irun.min() == 1
    assert irun.max() == N_ANGLES
    assert ien.min() == 1
    assert ien.max() == ENERGY_BIN_SIZE

    # We cannot check the values of idet directly because they are a non-contiguous
    # subset of all detector numbers. So just check that we have the correct number of
    # distinct detectors:
    assert np.unique(idet).size == N_DETECTORS * N_PIXELS_PER_DETECTOR



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Data for tests and documentation with BIFROST."""

from pathlib import Path

from ess.reduce.data import make_registry

_registry = make_registry(
    'ess/bifrost',
    files={
        "BIFROST_20240914T053723.h5": "md5:0f2fa5c9a851f8e3a4fa61defaa3752e",
        "computed_energy_data_simulated_5x2.h5": "md5:57408fa10aa4689c43630f994cff8d30",
        "BIFROST-simulation-tof-lookup-table.h5": "blake2b:682021920a355f789da37b18029719fe20569d86db26cdaf5f3d916d2f76f9360907960ba86903be4cab489d39f1b6f9f265f3a4ab3f82c5e095afa4a2c456af",  # noqa: E501
        "BIFROST-simulation-lookup-table.h5": "md5:6d776afa591d4a83c91ad0142bbfc53d",
    },
    version="7",
)


def get_path(name: str) -> Path:
    """
    Return the path to a data file bundled with ess.bifrost.

    This function only works with example data and cannot handle
    paths to custom files.
    """
    return _registry.get_path(name)


def simulated_elastic_incoherent_with_phonon() -> Path:
    """Simulated data for elastic incoherent scattering including a phonon."""
    return get_path("BIFROST_20240914T053723.h5")


def lookup_table_simulation() -> Path:
    """Wavelength lookup table for the simulated BIFROST data.

    This table was computed with `tof <https://github.com/scipp/tof>`_
    using `Create a wavelength lookup table for BIFROST
    <../../user-guide/bifrost/bifrost-make-wavelength-lookup-table.rst>`_
    with ``NumberOfSimulatedNeutrons = 5_000_000``.
    """
    return get_path("BIFROST-simulation-lookup-table.h5")


def computed_energy_data_simulated_5x2() -> Path:
    """Energy and momentum transfer computed from the simulated BIFROST data.

    This reference was computed with 10 detectors forming a 5x2 grid
    (arc=5, channel=2).
    """
    return get_path("computed_energy_data_simulated_5x2.h5")


__all__ = [
    "computed_energy_data_simulated_5x2",
    "get_path",
    "lookup_table_simulation",
    "simulated_elastic_incoherent_with_phonon",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""A single crystal reduction workflow for BIFROST's Bragg peak monitor."""

from .q_map import make_q_map
from .workflow import (
    BifrostBraggPeakMonitorWorkflow,
    BifrostSimulationBraggPeakMonitorWorkflow,
)

__all__ = [
    "BifrostBraggPeakMonitorWorkflow",
    "BifrostSimulationBraggPeakMonitorWorkflow",
    "make_q_map",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Domain types for single crystal diffraction on BIFROST."""

from dataclasses import dataclass
from typing import NewType

import sciline
import scipp as sc

from ess.spectroscopy.types import RunType


class CountsWithQMapCoords(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Counts with various Q coordinates.

    Has (event) coordinates:

    - 'Q_parallel': projection along the beam direction, aka 'Qz'
    - 'Q_perpendicular': perpendicular component of Q to the beam, aka 'Qx'
    - 'Q': absolute value of Q
    """


@dataclass(frozen=True, slots=True)
class QProjection:
    """Projection vectors in Q."""

    parallel: sc.Variable
    perpendicular: sc.Variable


QParallelBins = NewType("QParallelBins", sc.Variable)
"""Binning in Q_parallel (aka Qz)."""

QPerpendicularBins = NewType("QPerpendicularBins", sc.Variable)
"""Binning in Q_perpendicular (aka Qx)."""


class IntensityQparQperp(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Histogrammed data in Q_parallel and Q_perpendicular.

    Here, Q_parallel is the component of Q along the beam direction and Q_perpendicular
    is the non-zero component of Q perpendicular to the beam direction.
    For BIFROST, these are Qz and Qx respectively. (Qy is zero because the
    detector is in the x-z plane with the sample.)
    """


class IntensitySampleRotation(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Data integrated over |Q| and histogrammed in the sample rotation angle."""


QRange = NewType("QRange", tuple[sc.Variable, sc.Variable])
"""Range of |Q| to integrate over."""

SampleRotationBins = NewType("SampleRotationBins", sc.Variable)
"""Binning in the sample rotation angle."""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Utilities for computing real neutron time-of-flight."""

import scippnexus as snx

from ess.reduce import unwrap as reduce_unwrap
from ess.reduce.unwrap.types import DetectorLtotal
from ess.spectroscopy.types import (
    DataGroupedByRotation,
    ErrorLimitedLookupTable,
    PulseStrideOffset,
    RawDetector,
    RunType,
    WavelengthDetector,
)


def detector_wavelength_data(
    sample_data: DataGroupedByRotation[RunType],
    lookup: ErrorLimitedLookupTable[snx.NXdetector],
    ltotal: DetectorLtotal[RunType],
    pulse_stride_offset: PulseStrideOffset,
) -> WavelengthDetector[RunType]:
    """
    Convert the time-of-arrival data to wavelength data using a lookup table.

    The output data will have a wavelength coordinate.

    This is a wrapper around
    :func:`ess.reduce.unwrap.detector_wavelength_data`
    for different input types.
    """
    return reduce_unwrap.to_wavelength.detector_wavelength_data(
        detector_data=RawDetector[RunType](sample_data),
        lookup=lookup,
        ltotal=ltotal,
        pulse_stride_offset=pulse_stride_offset,
    )


providers = (detector_wavelength_data,)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Coordinate conversions for indirect spectroscopy."""

import numpy as np
import scipp as sc
import scipp.constants
import scippnexus as snx
from scippneutron.conversion.tof import (
    energy_from_wavelength,
    wavevector_from_wavelength,
)

from ..types import (
    EnergyQDetector,
    GravityVector,
    IncidentEnergyDetector,
    InelasticCoordTransformGraph,
    MonitorCoordTransformGraph,
    NormalizedIncidentEnergyDetector,
    Position,
    PrimarySpecCoordTransformGraph,
    RunType,
    SecondarySpecCoordTransformGraph,
    WavelengthDetector,
)


def incident_energy_from_wavelength(*, incident_wavelength: sc.Variable) -> sc.Variable:
    return energy_from_wavelength(wavelength=incident_wavelength)


def incident_wavevector_from_incident_wavelength(
    *, incident_wavelength: sc.Variable, incident_beam: sc.Variable
) -> sc.Variable:
    return wavevector_from_wavelength(
        wavelength=incident_wavelength, beam=incident_beam
    )


def energy_transfer(
    *, incident_energy: sc.Variable, final_energy: sc.Variable
) -> sc.Variable:
    r"""Compute the energy transfer.

    Here, the energy transfer is defined the same as in ScippNeutron
    (https://scipp.github.io/scippneutron/user-guide/coordinate-transformations.html)
    to be

    .. math::

        \Delta E = E_i - E_f

    Parameters
    ----------
    incident_energy:
        The neutron energy :math:`E_i` before scattering.
    final_energy:
        The neutron energy :math:`E_f` after scattering.

    Returns
    -------
    :
        The energy transfer :math:`\Delta E`.
    """
    return incident_energy - final_energy.to(unit=incident_energy.unit, copy=False)


def lab_momentum_transfer_from_wavevectors(
    incident_wavevector: sc.Variable, final_wavevector: sc.Variable
) -> sc.Variable:
    r"""Compute the momentum transfer in the lab frame.

    Here, the momentum transfer is defined as

    .. math::

        \vec{Q} = \vec{k_i} - \vec{k_f}

    Parameters
    ----------
    incident_wavevector:
        The neutron wavevector :math:`\vec{k_i}` before scattering.
    final_wavevector:
        The neutron wavevector :math:`\vec{k_f}` after scattering.

    Returns
    -------
    :
        The momentum transfer :math:`\vec{Q}` in the lab frame.
    """
    return (
        incident_wavevector.to(unit=final_wavevector.unit, copy=False)
        - final_wavevector
    )


def lab_momentum_transfer_from_incident_energy(
    incident_energy: sc.Variable,
    incident_beam: sc.Variable,
    final_wavevector: sc.Variable,
) -> sc.Variable:
    r"""Compute the momentum transfer in the lab frame.

    Here, the momentum transfer is defined as

    .. math::

        \vec{Q} = \vec{k_i} - \vec{k_f}

    with

    .. math::

        \vec{k_i} = \frac{2\pi}{h} \hat{b}_i \sqrt{2 M_n E_i}

    Parameters
    ----------
    incident_energy:
        The neutron energy :math:`E_i` before scattering.
    incident_beam:
        The vector of the incident beam in the lab frame :math:`\vec{b}_i`.
    final_wavevector:
        The neutron wavevector :math:`\vec{k_f}` after scattering.

    Returns
    -------
    :
        The momentum transfer :math:`\vec{Q}` in the lab frame.
    """
    # Typical shapes:
    #  - incident_beam: ()
    #  - incident_energy: (incident_energy,)  (small)
    #  - final_wavevector: (detector,)  (large-ish but 1D)
    #  - result: (detector, incident_energy)
    incident_wavevector = (
        (2 * np.pi / sc.constants.h)
        * (incident_beam / sc.norm(incident_beam))
        * sc.sqrt(2 * sc.constants.m_n * incident_energy)
    )
    # Order of operations such that the final result has dim order described above.
    return -final_wavevector + incident_wavevector.to(
        unit=final_wavevector.unit, copy=False
    )


def rotate_to_sample_table_momentum_transfer(
    *,
    a3: sc.Variable,
    lab_momentum_transfer: sc.Variable,
    gravity: sc.Variable,
) -> sc.Variable:
    r"""Rotate the momentum transfer vector into the sample-table coordinate system

    Here, the momentum transfer is defined as

    .. math::

        \vec{Q} = \vec{k_i} - \vec{k_f}

    Note
    ----
    When a3 is zero, the sample-table and lab coordinate systems are the same.
    That is, Z is along the incident beam, Y is opposite the gravitational force,
    and X completes the right-handed coordinate system. The sample-table angle, a3,
    has a rotation vector along Y, such that a positive 90-degree rotation places the
    sample-table Z along the lab X.

    Parameters
    ----------
    a3:
        The rotation angle of the sample table around the laboratory Y axis
    lab_momentum_transfer:
        The momentum transfer in the laboratory coordinate system
    gravity:
        The gravity vector which indicates the vertical axis.

    Returns
    -------
    :
        The momentum transfer in the sample-table coordinate system.
    """
    vertical = -gravity / sc.norm(gravity)
    # negative a3 since we rotate coordinates, not axes here
    return sc.spatial.rotations_from_rotvecs(-a3 * vertical) * lab_momentum_transfer


def inelastic_coordinate_transformation_graph_at_sample(
    gravity: GravityVector,
) -> InelasticCoordTransformGraph:
    return InelasticCoordTransformGraph(
        {
            'energy_transfer': energy_transfer,
            'incident_energy': incident_energy_from_wavelength,
            'incident_wavevector': incident_wavevector_from_incident_wavelength,
            'gravity': lambda: gravity,
            'lab_momentum_transfer': lab_momentum_transfer_from_wavevectors,
            'sample_table_momentum_transfer': rotate_to_sample_table_momentum_transfer,
        }
    )


def add_inelastic_coordinates(
    data: NormalizedIncidentEnergyDetector[RunType], graph: InelasticCoordTransformGraph
) -> EnergyQDetector[RunType]:
    transformed = data.transform_coords(
        [
            'energy_transfer',
            'sample_table_momentum_transfer',
            # These are inputs, but we want to preserve them
            'a3',
            'a4',
        ],
        graph=graph,
        keep_aliases=False,
        keep_inputs=False,
        keep_intermediate=False,
        rename_dims=False,
    )
    return EnergyQDetector[RunType](transformed)


def add_incident_energy(
    data: WavelengthDetector[RunType], graph: InelasticCoordTransformGraph
) -> IncidentEnergyDetector[RunType]:
    transformed = data.transform_coords(
        [
            'incident_energy',
            'incident_wavelength',
            # These are inputs, but we need them for binning:
            'a3',
            'a4',
            'final_energy',
            'final_wavevector',
        ],
        graph=graph,
        keep_aliases=False,
        keep_inputs=False,
        keep_intermediate=False,
        rename_dims=False,
    )
    return IncidentEnergyDetector[RunType](transformed)


def add_spectrometer_coords(
    data: sc.DataArray,
    primary_graph: PrimarySpecCoordTransformGraph[RunType],
    secondary_graph: SecondarySpecCoordTransformGraph[RunType],
) -> sc.DataArray:
    """Compute and add coordinates for the spectrometer.

    Parameters
    ----------
    data:
        Data array with beamline coordinates "position", "source_position", and
        "sample_position".
        Does not need to contain events or flight times.
    primary_graph:
        Coordinate transformation graph for the primary spectrometer.
    secondary_graph:
        Coordinate transformation graph for the secondary spectrometer.
        Must be a closure over analyzer parameters.
        And those parameters must have a compatible shape with ``data``.

    Returns
    -------
    :
        Input data with added spectrometer coordinates.
        This includes "final_energy", "secondary_flight_time", and "L1".
    """
    return data.transform_coords(
        (
            'final_energy',
            'final_wavevector',
            'incident_beam',
            'L1',
            'secondary_flight_time',
        ),
        graph={**primary_graph, **secondary_graph},
        keep_intermediate=False,
        keep_aliases=False,
        rename_dims=False,
    )


def monitor_coordinate_transformation_graph(
    source_position: Position[snx.NXsource, RunType],
) -> MonitorCoordTransformGraph[RunType]:
    from scippneutron.conversion.graph import beamline, tof

    return MonitorCoordTransformGraph[RunType](
        {
            **beamline.beamline(scatter=False),
            **tof.elastic_wavelength(start='tof'),
            'source_position': lambda: source_position,
        }
    )


providers = (
    add_inelastic_coordinates,
    add_incident_energy,
    inelastic_coordinate_transformation_graph_at_sample,
    monitor_coordinate_transformation_graph,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Cutting BIFROST data."""

from collections.abc import Callable

import scipp as sc

from ess.spectroscopy.types import (
    DataGroupedByRotation,
    InstrumentAngle,
    RawDetector,
    RunType,
    SampleAngle,
)


def group_by_rotation(
    data: RawDetector[RunType],
    sample_angle: SampleAngle[RunType],
    instrument_angle: InstrumentAngle[RunType],
) -> DataGroupedByRotation[RunType]:
    """Group data by rotation angles.

    Parameters
    ----------
    data:
        Detector events with time coordinates.
    sample_angle:
        Sample rotation angle "a3".
        May be time-dependent (1d array with dim "time") or scalar.
    instrument_angle:
        Instrument rotation angle "a4".
        May be time-dependent (1d array with dim "time") or scalar.

    Returns
    -------
    :
        ``data`` grouped by rotation angles "a3" and "a4".
    """
    graph = {
        'a3': _make_angle_from_time_calculator(sample_angle),
        'a4': _make_angle_from_time_calculator(instrument_angle),
    }
    grouped = data.transform_coords(('a3', 'a4'), graph=graph).group('a3', 'a4')
    return DataGroupedByRotation[RunType](grouped)


def _make_angle_from_time_calculator(angle: sc.DataArray) -> Callable[..., sc.Variable]:
    if angle.ndim == 0:
        return lambda: angle.data
    else:
        lut = sc.lookup(angle, 'time')
        return lambda event_time_zero: lut[event_time_zero]


providers = (group_by_rotation,)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import numpy as np
import pytest
import sciline
import scipp as sc
import scipp.testing
import scippnexus as snx

from ess import bifrost
from ess.bifrost.data import (
    computed_energy_data_simulated_5x2,
    lookup_table_simulation,
    simulated_elastic_incoherent_with_phonon,
)
from ess.spectroscopy.types import (
    EnergyQDetector,
    Filename,
    FrameMonitor3,
    LookupTableFilename,
    LookupTableRelativeErrorThreshold,
    NeXusDetectorName,
    RawDetector,
    RawMonitor,
    SampleRun,
    UncertaintyBroadcastMode,
    WavelengthMonitor,
)


@pytest.fixture(scope='module')
def simulation_detector_names() -> list[NeXusDetectorName]:
    with snx.File(simulated_elastic_incoherent_with_phonon()) as f:
        names = list(f['entry']['instrument'][snx.NXdetector].keys())
    return names[:10]  # First 10 detectors form a 5x2 grid (arc=5, channel=2)


@pytest.fixture
def workflow(simulation_detector_names: list[NeXusDetectorName]) -> sciline.Pipeline:
    workflow = bifrost.BifrostSimulationWorkflow(simulation_detector_names)
    workflow[Filename[SampleRun]] = simulated_elastic_incoherent_with_phonon()
    workflow[LookupTableFilename] = lookup_table_simulation()
    workflow[LookupTableRelativeErrorThreshold] = {
        'detector': np.inf,
        '110_frame_3': np.inf,
    }
    workflow[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.drop
    return workflow


def test_simulation_workflow_can_load_detector() -> None:
    workflow = bifrost.BifrostSimulationWorkflow(
        [NeXusDetectorName("125_channel_1_1_triplet")]
    )
    workflow[Filename[SampleRun]] = simulated_elastic_incoherent_with_phonon()
    results = sciline.compute_mapped(workflow, RawDetector[SampleRun])
    result = results.iloc[0]

    assert result.bins is not None
    assert set(result.dims) == {'tube', 'length'}
    assert result.sizes['tube'] == 3
    assert 'position' in result.coords


def test_simulation_workflow_can_load_monitor(workflow: sciline.Pipeline) -> None:
    result = workflow.compute(RawMonitor[SampleRun, FrameMonitor3])

    assert result.bins is None
    assert 'position' in result.coords


def test_simulation_workflow_can_compute_energy_data(
    workflow: sciline.Pipeline,
) -> None:
    energy_data = workflow.compute(EnergyQDetector[SampleRun])

    assert energy_data.sizes == {
        'arc': 5,
        'channel': 2,
        'tube': 3,
        'length': 100,
        'a3': 180,
        'a4': 1,
    }
    expected_coords = {'a3', 'a4', 'detector_number'}
    assert expected_coords.issubset(energy_data.coords)
    expected_event_coords = {
        'energy_transfer',
        'sample_table_momentum_transfer',
    }
    assert expected_event_coords.issubset(energy_data.bins.coords)

    # Check that conversions do not raise, i.e., units have expected dimensions.
    energy_data.coords['a3'].to(unit='rad')
    energy_data.coords['a4'].to(unit='rad')
    energy_data.bins.coords['energy_transfer'].to(unit='meV')
    energy_data.bins.coords['sample_table_momentum_transfer'].to(unit='1/Å')


def test_simulation_workflow_can_compute_wavelength_monitor(
    workflow: sciline.Pipeline,
) -> None:
    monitor = workflow.compute(WavelengthMonitor[SampleRun, FrameMonitor3])
    assert set(monitor.dims) == {'time', 'incident_wavelength'}
    expected_coords = {'position', 'incident_wavelength', 'time'}
    assert expected_coords.issubset(monitor.coords)
    assert monitor.bins is None


def test_simulation_workflow_produces_the_same_data_as_before(
    workflow: sciline.Pipeline,
) -> None:
    energy_data = workflow.compute(EnergyQDetector[SampleRun])
    sc.io.save_hdf5(energy_data, 'computed_energy_data_simulated_5x2.h5')
    expected = sc.io.load_hdf5(computed_energy_data_simulated_5x2())

    assert not energy_data.masks
    assert not energy_data.bins.masks

    assert set(expected.coords.keys()).issubset(set(energy_data.coords.keys()))
    for name in expected.coords.keys():
        sc.testing.assert_allclose(energy_data.coords[name], expected.coords[name])

    assert energy_data.bins.coords.keys() == expected.bins.coords.keys()
    for name in energy_data.bins.coords.keys():
        sc.testing.assert_allclose(
            energy_data.bins.coords[name], expected.bins.coords[name]
        )

    sc.testing.assert_allclose(energy_data.bins.data, expected.bins.data)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""SQW output for BIFROST."""

import dataclasses

import numpy as np
import numpy.typing as npt
import scipp as sc
from scippneutron.io import sqw
from scippneutron.metadata import ESS_SOURCE

from ess.spectroscopy.types import (
    Beamline,
    EnergyBins,
    GravityVector,
    Measurement,
    NormalizedIncidentEnergyDetector,
    OutFilename,
    PulsePeriod,
    SampleRun,
    SQWBinSizes,
)

# Projection axes
# For now, we always project onto the cartesian axes for simplicity.
_AXIS_U = sc.vector([1, 0, 0], unit="1/angstrom")
_AXIS_V = sc.vector([0, 1, 0], unit="1/angstrom")
_AXIS_W = sc.vector([0, 0, 1], unit="1/angstrom")


def save_sqw(
    path: OutFilename,
    events: NormalizedIncidentEnergyDetector[SampleRun],
    *,
    bin_sizes: SQWBinSizes,
    energy_bins: EnergyBins,
    beamline: Beamline[SampleRun],
    measurement: Measurement[SampleRun],
    sample: sqw.SqwIXSample,
    pulse_period: PulsePeriod,
    gravity: GravityVector,
) -> None:
    """Save events recorded at BIFROST to an SQW file.

    The files written by this function can be read by
    `Horace <https://pace-neutrons.github.io/Horace>`_ for further processing.
    For details on the file format, see the documentation for Horace as well as
    the `SQW <https://scipp.github.io/scippneutron/developer/file-formats/sqw.html>`_
    developer guide in ScippNeutron.

    The input to this function should not yet contain inelastic coordinates like
    'energy_transfer' or 'sample_table_momentum_transfer'.
    Instead, ``save_sqw`` histograms the events according to detector number,
    experiment setting, and incident energy and computes inelastic
    coordinates on that histogram.
    This is done for consistency with Horace.

    The data is encoded as a 4-dimensional array, and Horace calls the dimensions
    ``u1 ... u4``. The first three correspond to the three dimensions of $\\vec{Q}$
    and the ``u4`` corresponds to energy transfer. Horace supports customizable
    projections onto the cardinal axes $Q_x, Q_y, Q_z$, but the implementation here
    has hard-coded ``(Qx, Qy, Qz) = (u1, u2, u3)``, however, you need to express the
    binning in terms of the `u`'s.

    Note
    ----
    This function requires large amounts of memory to construct intermediate
    arrays and fill those with null observations.

    Parameters
    ----------
    path:
        Path of the file to write.
    events:
        Binned data array with shape ``(*logical, "a3", "a4")`` where ``logical``
        is any number of logical detector dimensions (e.g., arc, tube).
        Must have an ``"incident_energy"`` event coordinate.
    bin_sizes:
        Sizes of the output 'image' bins.
        The data will be ordered based on these bins, but it is possible
        to change the binning later in Horace.
        E.g., ``{'u1': 50, 'u2': 50, 'u3': 50, 'u4': 50}``.
    energy_bins:
        Bin edges or number of bins for (incident) energy.
        The events are histogrammed according to these bins, so this determines the
        final energy resolution that is available in Horace.
        Should typically be larger than ``bin_sizes['energy_transfer']``.
    beamline:
        Beamline metadata.
    measurement:
        Measurement metadata.
    sample:
        Sample metadata in SQW format.
    pulse_period:
        Pulse period of the neutron source.
    gravity:
        Vector indicating the direction of gravity.
        Used to define the lab coordinate system.
        See also
        :func:`ess.spectroscopy.indirect.conversion.rotate_to_sample_table_momentum_transfer`.

    See also
    --------
    scippneutron.io.sqw:
        For low-level SQW I/O and the underlying implementation of ``save_sqw``.
    """
    if np.unique(events.coords['a4'].values).size != 1:
        # We need to support this eventually, but we don't
        # have data for a moving detector vessel yet.
        raise NotImplementedError("a4 must be constant for all events")

    flat_events = _flatten_events(events)
    del events  # 'move' events into _flatten_events
    _filter_and_convert_coords_in_place(flat_events)

    sample_angle = flat_events.coords['a3']

    observations = _histogram_detector_setting_ei(flat_events, energy_bins=energy_bins)
    del flat_events  # 'move' flat_events into _histogram_detector_setting_ei
    final_energy = observations.coords['final_energy']
    observations = _with_inelastic_coords(observations, gravity)
    energy_transfer = observations.coords['energy_transfer'].rename_dims(
        incident_energy='energy_transfer'
    )
    observations = observations.flatten(to="observation")

    dnd, counts, binned_observations = _bin_image(observations, bin_sizes)
    del observations  # 'move' observations into _bin_image
    pix = _make_pixel_data(binned_observations)
    del binned_observations  # 'move' binned_observations into _make_pixel_data
    pix_buffer = _make_pixel_buffer(pix)
    del pix  # 'move' pix into _make_pixel_buffer

    experiments = _make_experiments(
        energy_transfer=energy_transfer,
        final_energy=final_energy,
        sample_angle=sample_angle,
    )
    builder = (
        sqw.Sqw.build(path, title=measurement.title)
        .add_default_instrument(_make_instrument(beamline, pulse_period))
        .add_default_sample(sample)
        .add_dnd_data(_make_dnd_metadata(dnd, sample), data=dnd.data, counts=counts)
        .add_pixel_data(pix_buffer, experiments=experiments)
    )
    builder.create()


def _flatten_events(
    events: NormalizedIncidentEnergyDetector[SampleRun],
) -> sc.DataArray:
    """Flatten events from (*logical, a3, a4) to (detector, setting).

    Also assigns the 'irun' coordinate as that requires the initial shape.
    """
    logical_dims = [dim for dim in events.dims if dim not in ("a3", "a4")]
    aux = events.flatten(logical_dims, 'detector')

    n_a3 = aux.sizes['a3']
    aux.coords['i_a3'] = sc.arange('a3', n_a3, dtype='float32', unit=None)
    aux.coords['i_a4'] = sc.arange('a4', aux.sizes['a4'], dtype='float32', unit=None)
    flat = aux.flatten(['a3', 'a4'], 'setting')
    return flat.assign_coords(
        irun=flat.coords.pop('i_a3')
        + flat.coords.pop('i_a4') * sc.index(n_a3)
        + sc.index(1, dtype='float32')  # 1-based indexing
    )


def _filter_and_convert_coords_in_place(flat_events: sc.DataArray) -> None:
    """Filter out all coords we no longer need and convert remaining coords to float32.

    This function mainly serves to reduce memory usage.
    The conversion to float32 is ultimately required when building the 'pixel data',
    we do it here to reduce size early. This should not have a significant impact
    on precision as we do not reduce these coordinates.
    The event weights remain in float64 or int64 for now because
    they do need to be summed.
    """
    flat_events.coords['final_energy'] = flat_events.coords['final_energy'].to(
        dtype='float32', copy=False
    )
    flat_events.bins.coords['incident_energy'] = flat_events.bins.coords[
        'incident_energy'
    ].to(dtype='float32', copy=False)

    flat_events.coords['idet'] = flat_events.coords.pop('detector_number').to(
        dtype='float32', copy=False
    )
    # to 1-based indexing:
    flat_events.coords['idet'] += sc.scalar(1.0, dtype='float32', unit=None)

    keep = {
        'a3',
        'final_energy',
        'final_wavevector',
        'gravity',
        'idet',
        'incident_beam',
        'irun',
    }
    for dim in set(flat_events.coords) - keep:
        del flat_events.coords[dim]


def _histogram_detector_setting_ei(
    flat_events: sc.DataArray, *, energy_bins: EnergyBins
) -> sc.DataArray:
    """Histogram the events in (detector, setting, incident_energy).

    This also converts the weights to float32 as no further reductions
    will be performed for the 'pixel data'.
    We accept a loss of precision for the 'image data' here which will be
    converted back to float64 later.
    """
    hist = flat_events.hist(incident_energy=energy_bins)
    hist.coords['incident_energy'] = sc.midpoints(hist.coords['incident_energy']).to(
        dtype='float32', copy=False
    )
    hist.coords['ien'] = sc.arange(
        'incident_energy',
        1,  # 1-based indexing
        hist.sizes['incident_energy'] + 1,
        dtype='float32',
        unit=None,
    )

    hist.data = hist.data.to(dtype='float32', copy=False)

    return hist


def _with_inelastic_coords(
    observations: sc.DataArray, gravity: GravityVector
) -> sc.DataArray:
    """Compute and assign Qx, Qy, Qz, and energy_transfer.

    This also drops all coordinates that are no longer needed.
    """
    from ess.spectroscopy.indirect.conversion import (
        energy_transfer,
        lab_momentum_transfer_from_incident_energy,
        rotate_to_sample_table_momentum_transfer,
    )

    graph = {
        'lab_momentum_transfer': lab_momentum_transfer_from_incident_energy,
        'sample_table_momentum_transfer': rotate_to_sample_table_momentum_transfer,
        'energy_transfer': energy_transfer,
        'gravity': lambda: gravity,
    }
    aux = observations.transform_coords(
        ['sample_table_momentum_transfer', 'energy_transfer'],
        graph=graph,
        keep_inputs=False,
        keep_intermediate=False,
        quiet=True,
    )
    q = aux.coords.pop('sample_table_momentum_transfer')

    return aux.assign_coords(
        energy_transfer=aux.coords.pop('energy_transfer').to(
            dtype='float32', copy=False
        ),
        **{
            f'Q{dim}': _project_onto(q, axis).to(dtype='float32')
            for dim, axis in zip('xyz', (_AXIS_U, _AXIS_V, _AXIS_W), strict=True)
        },
    )


def _project_onto(vec: sc.Variable, axis: sc.Variable) -> sc.Variable:
    axis = axis / sc.norm(axis)
    # Optimization to avoid huge dot-products:
    match list(axis.values):
        case [1, 0, 0]:
            return vec.fields.x
        case [0, 1, 0]:
            return vec.fields.y
        case [0, 0, 1]:
            return vec.fields.z
        case _:
            return sc.dot(axis, vec)


def _bin_image(
    observations: sc.DataArray, bin_sizes: SQWBinSizes
) -> tuple[sc.DataArray, sc.Variable, sc.DataArray]:
    # Dim order to match pixel data (note the transpose for dnd)
    dim_order = {'energy_transfer': 'u4', 'Qz': 'u3', 'Qy': 'u2', 'Qx': 'u1'}
    binned = observations.bin(
        {coord: bin_sizes[dim] for coord, dim in dim_order.items()}
    ).rename(energy_transfer='u4', Qz='u3', Qy='u2', Qx='u1')

    image = binned.transpose(['u1', 'u2', 'u3', 'u4'])
    # The dense DND array contains the averages of the observations in `image`.
    # Note that `image` contains 'observations', not 'events', so the expectation value
    # is the arithmetic mean, not the sum.
    # See also https://scipp.github.io/scippneutron/developer/file-formats/sqw.html#dnd-data-blocks
    dnd = image.bins.mean().to(dtype='float64', copy=False)
    counts = image.bins.size()

    return dnd, counts, binned


def _make_pixel_data(binned_observations: sc.DataArray) -> sc.DataArray:
    pix = binned_observations.data.bins.concat().value
    del binned_observations

    # Convert to expected units:
    return sc.DataArray(
        pix.data,
        coords={
            'u1': pix.coords['u1'].to(unit='1/Å', copy=False),
            'u2': pix.coords['u2'].to(unit='1/Å', copy=False),
            'u3': pix.coords['u3'].to(unit='1/Å', copy=False),
            'u4': pix.coords['u4'].to(unit='meV', copy=False),
            'idet': pix.coords['idet'],
            'irun': pix.coords['irun'],
            'ien': pix.coords['ien'],
        },
    )


def _make_pixel_buffer(pix: sc.DataArray) -> npt.NDArray[np.float32]:
    return np.c_[
        *(
            pix.coords[name].values
            for name in ('u1', 'u2', 'u3', 'u4', 'irun', 'idet', 'ien')
        ),
        pix.values,
        sc.variances(pix).values,
    ]


def _make_instrument(
    beamline: Beamline, pulse_period: PulsePeriod
) -> sqw.SqwIXNullInstrument:
    return sqw.SqwIXNullInstrument(
        name=beamline.name,
        source=sqw.SqwIXSource(
            name="ESS",
            target_name=ESS_SOURCE.name,
            frequency=1 / pulse_period,
        ),
    )


def _make_dnd_metadata(
    dnd: sc.DataArray, sample: sqw.SqwIXSample
) -> sqw.SqwDndMetadata:
    """Create the DND metadata for an SQW file.

    Parameters
    ----------
    dnd:
        DND 'image data'.
        Must be histogrammed along the u-axes with linspace coordinates.
    sample:
        Sample metadata.
    """
    img_range = [
        sc.concat([dnd.coords[name].min(), dnd.coords[name].max()], dim=name)
        for name in ("u1", "u2", "u3", "u4")
    ]

    n_bins_all_dims = sc.array(
        dims=["axis"],
        values=[dnd.sizes[name] for name in ("u1", "u2", "u3", "u4")],
        unit=None,
    )

    return sqw.SqwDndMetadata(
        axes=sqw.SqwLineAxes(
            title="Instrument Axes",
            label=["Qx", "Qy", "Qz", "Delta E"],
            img_scales=[
                sc.scalar(1.0, unit="1/angstrom"),
                sc.scalar(1.0, unit="1/angstrom"),
                sc.scalar(1.0, unit="1/angstrom"),
                sc.scalar(1.0, unit="meV"),
            ],
            img_range=img_range,
            n_bins_all_dims=n_bins_all_dims,
            single_bin_defines_iax=sc.array(dims=["axis"], values=[True] * 4),
            dax=sc.arange("axis", 4, unit=None),
            offset=[
                sc.scalar(0.0, unit="1/angstrom"),
                sc.scalar(0.0, unit="1/angstrom"),
                sc.scalar(0.0, unit="1/angstrom"),
                sc.scalar(0.0, unit="meV"),
            ],
            changes_aspect_ratio=True,
        ),
        proj=sqw.SqwLineProj(
            title="Null Projection",
            lattice_spacing=sample.lattice_spacing,
            lattice_angle=sample.lattice_angle,
            offset=[
                sc.scalar(0.0, unit="1/angstrom"),
                sc.scalar(0.0, unit="1/angstrom"),
                sc.scalar(0.0, unit="1/angstrom"),
                sc.scalar(0.0, unit="meV"),
            ],
            label=["Qx", "Qy", "Qz", "Delta E"],
            u=_AXIS_U,
            v=_AXIS_V,
            w=_AXIS_W,
            non_orthogonal=False,
            type="aaa",
        ),
    )


def _make_experiments(
    *,
    energy_transfer: sc.Variable,
    final_energy: sc.Variable,
    sample_angle: sc.Variable,
) -> list[sqw.SqwIXExperiment]:
    experiment_template = sqw.SqwIXExperiment(
        run_id=0,  # converted to 1-based by ScippNeutron
        efix=final_energy,
        emode=sqw.EnergyMode.indirect,
        en=energy_transfer,
        psi=sc.scalar(0.0, unit="rad"),
        u=_AXIS_U,
        v=_AXIS_V,
        omega=sc.scalar(0.0, unit="rad"),
        dpsi=sc.scalar(0.0, unit="rad"),
        gl=sc.scalar(0.0, unit="rad"),
        gs=sc.scalar(0.0, unit="rad"),
    )
    return [
        dataclasses.replace(experiment_template, run_id=i, psi=a3)
        for i, a3 in enumerate(sample_angle)
    ]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""BIFROST Bragg peak monitor workflows."""

import sciline

from ess.reduce import unwrap as reduce_unwrap
from ess.spectroscopy.indirect.time_of_flight import TofWorkflow
from ess.spectroscopy.types import (
    FrameMonitor1,
    FrameMonitor2,
    FrameMonitor3,
    SampleRun,
)

from ..cutting import group_by_rotation
from ..io import nexus
from ..io.mcstas import convert_simulated_time_to_event_time_offset
from ..workflow import default_parameters, simulation_default_parameters
from . import conversion, q_map, time_of_flight

_PROVIDERS = (
    *nexus.providers,
    *conversion.providers,
    *q_map.providers,
    *time_of_flight.providers,
    group_by_rotation,
)

_SIMULATION_PROVIDERS = (
    *nexus.providers,
    *conversion.providers,
    *q_map.providers,
    *time_of_flight.providers,
    convert_simulated_time_to_event_time_offset,
    group_by_rotation,
)


def BifrostBraggPeakMonitorWorkflow() -> sciline.Pipeline:
    workflow = TofWorkflow(
        run_types=(SampleRun,),
        monitor_types=(FrameMonitor1, FrameMonitor2, FrameMonitor3),
    )
    # Use the vanilla implementation instead of the indirect geometry one:
    workflow.insert(reduce_unwrap.to_wavelength.detector_wavelength_data)
    for provider in _PROVIDERS:
        workflow.insert(provider)
    for key, val in default_parameters().items():
        workflow[key] = val
    return workflow


def BifrostSimulationBraggPeakMonitorWorkflow() -> sciline.Pipeline:
    workflow = TofWorkflow(
        run_types=(SampleRun,),
        monitor_types=(FrameMonitor1, FrameMonitor2, FrameMonitor3),
    )
    # Use the vanilla implementation instead of the indirect geometry one:
    workflow.insert(reduce_unwrap.to_wavelength.detector_wavelength_data)
    for provider in _SIMULATION_PROVIDERS:
        workflow.insert(provider)
    for key, val in simulation_default_parameters().items():
        workflow[key] = val
    return workflow



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Build a Q-map for single crystal diffraction."""

from collections.abc import Callable
from typing import Any

import numpy as np
import numpy.typing as npt
import plopp as pp
import scipp as sc
from matplotlib.axes import Axes

from ess.spectroscopy.types import QDetector, RunType

from .types import (
    CountsWithQMapCoords,
    IntensityQparQperp,
    IntensitySampleRotation,
    QParallelBins,
    QPerpendicularBins,
    QProjection,
    QRange,
    SampleRotationBins,
)


def project_momentum_transfer(
    counts: QDetector[RunType],
    *,
    q_projection: QProjection,
) -> CountsWithQMapCoords[RunType]:
    """Project Q onto given axes and also compute the norm of Q."""
    transformed = counts.transform_coords(
        ['Q_parallel', 'Q_perpendicular', 'Q'],
        graph={
            'Q_parallel': _make_projection_kernel(q_projection.parallel),
            'Q_perpendicular': _make_projection_kernel(q_projection.perpendicular),
            'Q': lambda sample_table_momentum_transfer: sc.norm(
                sample_table_momentum_transfer
            ),
        },
        keep_inputs=False,
    )
    if transformed.bins is not None:
        transformed.bins.coords['a3'] = sc.bins_like(
            transformed, transformed.coords['a3']
        )
        transformed = transformed.bins.concat()

    return CountsWithQMapCoords[RunType](transformed)


def _make_projection_kernel(vector: sc.Variable) -> Callable[..., sc.Variable]:
    normal = vector / sc.norm(vector)

    def projection(sample_table_momentum_transfer: sc.Variable) -> sc.Variable:
        return sc.dot(sample_table_momentum_transfer, normal)

    return projection


def default_q_projection() -> QProjection:
    return QProjection(
        parallel=sc.vector(value=[0, 0, 1]), perpendicular=sc.vector(value=[1, 0, 0])
    )


def histogram_qparallel_qperpendicular(
    events: CountsWithQMapCoords[RunType],
    q_parallel_bins: QParallelBins,
    q_perpendicular_bins: QPerpendicularBins,
) -> IntensityQparQperp[RunType]:
    """Histogram the data in Q_parallel and Q_perpendicular"""
    return IntensityQparQperp[RunType](
        events.hist(Q_perpendicular=q_perpendicular_bins, Q_parallel=q_parallel_bins)
    )


def integrate_q(
    events: CountsWithQMapCoords[RunType],
    q_range: QRange,
    sample_rotation_bins: SampleRotationBins,
) -> IntensitySampleRotation[RunType]:
    """Integrate the data over |Q| and histogram in the sample rotation."""
    unit = events.bins.coords['Q'].unit
    lo = q_range[0].to(unit=unit)
    hi = q_range[1].to(unit=unit)
    sliced = events.bins['Q', lo:hi]
    try:
        return IntensitySampleRotation[RunType](sliced.hist(a3=sample_rotation_bins))
    except ValueError as err:
        if "empty data range" in err.args[0].lower():
            # This happens when the user selects an empty range Q range,
            # which is mainly a problem in the widget.
            return IntensitySampleRotation[RunType](_empty_angle_array_like(sliced))
        raise


def make_q_map(
    events: sc.DataArray,
    q_parallel_bins: int | sc.Variable,
    q_perpendicular_bins: int | sc.Variable,
    sample_rotation_bins: int | sc.Variable,
    *,
    circle_n_points: int = 100,
    roi_fill_alpha: float = 0.2,
    roi_fill_color: Any = 'C1',
    roi_line_color: Any = 'C1',
) -> Any:
    """Build a figure with a 2D Q-map and a 1D slice for a range of Q values."""
    import ipywidgets as ipw
    from plopp.widgets import Box

    def make_q_hist(da: sc.DataArray) -> sc.DataArray:
        return histogram_qparallel_qperpendicular(
            da, q_parallel_bins, q_perpendicular_bins
        ).rename(Q_perpendicular=r'$Q_{\perp}$', Q_parallel='$Q_{||}$')

    def make_q_slice(da: sc.DataArray, q_range: tuple[float, float]) -> sc.DataArray:
        unit = da.bins.coords['Q'].unit
        lo = q_range[0] * unit
        hi = q_range[1] * unit
        return integrate_q(da, (lo, hi), sample_rotation_bins)

    q_lo = events.bins.coords['Q'].min().value
    q_hi = events.bins.coords['Q'].max().value
    slider = ipw.FloatRangeSlider(
        value=(q_lo, q_hi), min=q_lo, max=q_hi, step=(q_hi - q_lo) / 100
    )

    slider_node = pp.widget_node(slider)
    input_node = pp.Node(events)
    q_map_node = pp.Node(make_q_hist, input_node)
    q_slice_node = pp.Node(make_q_slice, input_node, slider_node)
    q_map_fig = pp.imagefigure(q_map_node, aspect='equal', norm='log')
    q_slice_fig = pp.linefigure(q_slice_node)

    roi_path = _ROICirclePath(q_lo, q_hi, circle_n_points)
    roi_circle = _ROICircle(
        roi_path,
        q_map_fig.ax,
        fill_alpha=roi_fill_alpha,
        fill_color=roi_fill_color,
        line_color=roi_line_color,
    )

    def update_roi(q_range: tuple[float, float]) -> None:
        roi_circle.set(*q_range)
        q_map_fig.canvas.draw()

    # Create a view without any output so that `update_roi` gets called
    # when the slider is moved. See https://github.com/scipp/plopp/pull/496
    pp.View(pp.Node(update_roi, slider_node))
    return Box([[q_map_fig], [q_slice_fig], [slider]])


class _ROICirclePath:
    # The path is encoded as an array xy with shape (n_points * 2 + 1, 2).
    # Where xy[:, 0] is the x coordinate and xy[:, 1] is the y coordinate.
    # xy[:n_points] is the outer circle, xy[n_points:-1] is the inner circle.
    # xy[-1] is the same as xy[0]; we close the path to be compatible with MPL.

    def __init__(self, r_inner: float, r_outer: float, n_points: int) -> None:
        self._r_inner = r_inner
        self._r_outer = r_outer
        self._n_points = n_points

        self._angles = np.linspace(0, 2 * np.pi, n_points, endpoint=True)
        # 2*n_points for the inner and outer circle,
        # +1 to repeat the first point at the end.
        self._xy = np.zeros((n_points * 2 + 1, 2), dtype=float)

        self.set_inner(r_inner)
        self.set_outer(r_outer)

    @property
    def r_inner(self) -> float:
        return self._r_inner

    @property
    def r_outer(self) -> float:
        return self._r_outer

    @property
    def closed_xy(self) -> npt.NDArray[float]:
        # includes the last point
        return self._xy

    @property
    def open_xy(self) -> npt.NDArray[float]:
        # does not include the last point
        return self._xy[:-1]

    @property
    def inner(self) -> npt.NDArray[float]:
        # does not include the last point
        return self._xy[self._n_points : -1]

    @property
    def outer(self) -> npt.NDArray[float]:
        # does not include the last point
        return self._xy[: self._n_points]

    def set_inner(self, r: float) -> None:
        # reverse=True makes MPL render this circle as a cutout.
        self._r_inner = r
        x, y = self.inner.T
        self._set_circle(x, y, r, reverse=True)

    def set_outer(self, r: float) -> None:
        self._r_outer = r
        x, y = self.outer.T
        self._set_circle(x, y, r, reverse=False)
        self._xy[-1] = self._xy[0]

    def _set_circle(
        self, x: npt.NDArray[float], y: npt.NDArray[float], r: float, reverse: bool
    ) -> None:
        if reverse:
            angles = self._angles[::-1]
        else:
            angles = self._angles

        np.cos(angles, out=x)
        x *= r

        np.sin(angles, out=y)
        y *= r


class _ROICircle:
    def __init__(
        self,
        path: _ROICirclePath,
        ax: Axes,
        *,
        fill_alpha: float = 0.2,
        fill_color: Any = 'C1',
        line_color: Any = 'C1',
    ) -> None:
        self._path = path
        self._fill = ax.fill(
            *self._path.open_xy.T, fill_color, alpha=fill_alpha, zorder=-1
        )[0]
        self._inner = ax.plot(*self._path.inner.T, c=line_color, zorder=2)[0]
        self._outer = ax.plot(*self._path.outer.T, c=line_color, zorder=2)[0]

    def set(self, r_inner: float, r_outer: float) -> None:
        if r_inner != self._path.r_inner:
            self._path.set_inner(r_inner)
            self._inner.set_data(*self._path.inner.T)
        if r_outer != self._path.r_outer:
            self._path.set_outer(r_outer)
            self._outer.set_data(*self._path.outer.T)
        self._fill.set_xy(self._path.closed_xy)


def _empty_angle_array_like(
    reference: sc.DataArray, *, dim: str = 'sample_rotation'
) -> sc.DataArray:
    a3_lo = reference.bins.coords[dim].min()
    a3_hi = reference.bins.coords[dim].max()
    return sc.DataArray(
        sc.array(
            dims=[dim],
            values=[0, 0],
            variances=[0, 0],
            unit=reference.unit,
            dtype=reference.dtype,
        ),
        coords={
            dim: sc.array(
                dims=[dim], values=[a3_lo.value, a3_hi.value], unit=a3_lo.unit
            )
        },
    )


providers = (
    default_q_projection,
    integrate_q,
    histogram_qparallel_qperpendicular,
    project_momentum_transfer,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Domain types for BIFROST.

This module supplements :mod:`ess.spectroscopy.types` with BIFROST-specific types.
"""

from typing import NewType

import sciline
import scipp as sc

from ess.spectroscopy.types import RunType


class ArcNumber(sciline.Scope[RunType, sc.Variable], sc.Variable): ...


class McStasRawDetector(sciline.Scope[RunType, sc.DataArray], sc.DataArray): ...


ArcEnergy = NewType('ArcEnergy', sc.Variable)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""BIFROST workflows."""

from typing import Any

import sciline
import scipp as sc
from scippnexus import NXdetector

from ess.spectroscopy.indirect.conversion import providers as conversion_providers
from ess.spectroscopy.indirect.kf import providers as kf_providers
from ess.spectroscopy.indirect.ki import providers as ki_providers
from ess.spectroscopy.indirect.time_of_flight import TofWorkflow
from ess.spectroscopy.types import (
    EmptyDetector,
    FrameMonitor0,
    FrameMonitor1,
    FrameMonitor2,
    FrameMonitor3,
    NeXusData,
    NeXusDetectorName,
    NeXusMonitorName,
    ProtonCharge,
    PulsePeriod,
    RawDetector,
    SampleRun,
    UncertaintyBroadcastMode,
)

from .cutting import providers as cutting_providers
from .detector import merge_triplets
from .detector import providers as detector_providers
from .io import mcstas, nexus
from .normalization import providers as normalisation_providers


def default_parameters() -> dict[type, Any]:
    """Default parameters for BifrostWorkflow."""
    return {
        NeXusMonitorName[FrameMonitor1]: '090_frame_1',
        NeXusMonitorName[FrameMonitor2]: '097_frame_2',
        NeXusMonitorName[FrameMonitor3]: '110_frame_3',
        PulsePeriod: 1.0 / sc.scalar(14.0, unit="Hz"),
        UncertaintyBroadcastMode: UncertaintyBroadcastMode.fail,
    }


def simulation_default_parameters() -> dict[type, Any]:
    """Default parameters for BifrostSimulationWorkflow."""
    return {
        NeXusMonitorName[FrameMonitor1]: '090_frame_1',
        NeXusMonitorName[FrameMonitor2]: '097_frame_2',
        NeXusMonitorName[FrameMonitor3]: '110_frame_3',
        PulsePeriod: 1.0 / sc.scalar(14.0, unit="Hz"),
        ProtonCharge[SampleRun]: sc.DataArray(sc.scalar(1.0, unit='pC')),
        UncertaintyBroadcastMode: UncertaintyBroadcastMode.fail,
    }


_PROVIDERS = (
    *nexus.providers,
    *conversion_providers,
    *detector_providers,
    *cutting_providers,
    *ki_providers,
    *kf_providers,
    *normalisation_providers,
)

_SIMULATION_PROVIDERS = (
    *nexus.providers,
    *conversion_providers,
    *detector_providers,
    *mcstas.providers,
    *cutting_providers,
    *ki_providers,
    *kf_providers,
    *normalisation_providers,
)


def BifrostSimulationWorkflow(
    detector_names: list[NeXusDetectorName],
) -> sciline.Pipeline:
    """Data reduction workflow for simulated BIFROST data.

    Parameters
    ----------
    detector_names:
        Names of ``NXdetector`` groups in the input NeXus file.

    Returns
    -------
    :
        A pipeline for reducing simulated BIFROST data.
    """
    workflow = TofWorkflow(
        run_types=(SampleRun,),
        monitor_types=(FrameMonitor0, FrameMonitor1, FrameMonitor2, FrameMonitor3),
    )
    for provider in _SIMULATION_PROVIDERS:
        workflow.insert(provider)
    for key, val in simulation_default_parameters().items():
        workflow[key] = val

    workflow[RawDetector[SampleRun]] = (
        workflow[RawDetector[SampleRun]]
        .map(_make_detector_name_mapping(detector_names))
        .reduce(func=merge_triplets)
    )

    return workflow


def BifrostWorkflow(
    detector_names: list[NeXusDetectorName],
) -> sciline.Pipeline:
    """Data reduction workflow for BIFROST."""
    workflow = TofWorkflow(
        run_types=(SampleRun,),
        monitor_types=(FrameMonitor0, FrameMonitor1, FrameMonitor2, FrameMonitor3),
    )
    for provider in _PROVIDERS:
        workflow.insert(provider)
    for key, val in default_parameters().items():
        workflow[key] = val

    workflow[EmptyDetector[SampleRun]] = (
        workflow[EmptyDetector[SampleRun]]
        .map(_make_detector_name_mapping(detector_names))
        .reduce(func=merge_triplets)
    )

    workflow[NeXusData[NXdetector, SampleRun]] = (
        workflow[NeXusData[NXdetector, SampleRun]]
        .map(_make_detector_name_mapping(detector_names))
        .reduce(func=concat_event_lists)
    )

    return workflow


# TODO remove or move
def concat_event_lists(
    *data: sc.DataArray,
) -> sc.DataArray:
    """Concatenate binned event lists into a single data array in 'event_time_zero'.

    Note that the output will likely have repeated values for 'event_time_zero'.
    E.g., if input 1 has times `[0, 1, 2]` and input 2 has times `[0, 2, 3]`,
    the output will have times `[0, 1, 2, 0, 2, 3]`.
    Note that this sawtooth pattern will disappear again after grouping into pixels.
    Preserving it will likely actually lead to more efficient memory
    access patterns when grouping.

    Parameters
    ----------
    data:
        Data arrays to concatenate.
        Must be binned in 'event_time_zero'.

    Returns
    -------
    :
        Concatenated data array.
    """
    return sc.concat(data, dim="event_time_zero")


def _make_detector_name_mapping(detector_names: list[NeXusDetectorName]) -> Any:
    # Use Pandas if possible to label the index.
    try:
        import pandas

        return pandas.DataFrame({NeXusDetectorName: detector_names}).rename_axis(
            index='triplet'
        )
    except ModuleNotFoundError:
        return {NeXusDetectorName: detector_names}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402, F401, I

"""Data reduction for ESS spectroscopy experiments.

This package contains common functionality for spectrometers at ESS.
For instrument-specifics, see

- :mod:`ess.bifrost`
"""

import importlib.metadata

try:
    __version__ = importlib.metadata.version("essspectroscopy")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Domain types for spectroscopy."""

from typing import Any, NewType, TypeVar

import sciline
import scipp as sc

from ess.reduce import unwrap
from ess.reduce.nexus import types as reduce_t
from ess.reduce.uncertainty import UncertaintyBroadcastMode as _UncertaintyBroadcastMode

# NeXus types

Beamline = reduce_t.Beamline
DetectorPositionOffset = reduce_t.DetectorPositionOffset
EmptyDetector = reduce_t.EmptyDetector
Filename = reduce_t.Filename
GravityVector = reduce_t.GravityVector
Measurement = reduce_t.Measurement
NeXusClass = reduce_t.NeXusClass
NeXusComponentLocationSpec = reduce_t.NeXusComponentLocationSpec
NeXusComponent = reduce_t.NeXusComponent
NeXusData = reduce_t.NeXusData
NeXusDetectorName = reduce_t.NeXusDetectorName
NeXusFileSpec = reduce_t.NeXusFileSpec
NeXusMonitorName = reduce_t.NeXusName
NeXusTransformation = reduce_t.NeXusTransformation
Position = reduce_t.Position
PreopenNeXusFile = reduce_t.PreopenNeXusFile
RawDetector = reduce_t.RawDetector
RawMonitor = reduce_t.RawMonitor

SampleRun = reduce_t.SampleRun
VanadiumRun = reduce_t.VanadiumRun

FrameMonitor0 = reduce_t.FrameMonitor0
FrameMonitor1 = reduce_t.FrameMonitor1
FrameMonitor2 = reduce_t.FrameMonitor2
FrameMonitor3 = reduce_t.FrameMonitor3

# Type vars

RunType = TypeVar("RunType", SampleRun, VanadiumRun)
MonitorType = TypeVar(
    "MonitorType",
    FrameMonitor0,
    FrameMonitor1,
    FrameMonitor2,
    FrameMonitor3,
)

# Time-of-flight types

WavelengthDetector = unwrap.WavelengthDetector
WavelengthMonitor = unwrap.WavelengthMonitor
LookupTableRelativeErrorThreshold = unwrap.LookupTableRelativeErrorThreshold
MonitorLtotal = unwrap.MonitorLtotal
PulseStride = unwrap.PulseStride
PulseStrideOffset = unwrap.PulseStrideOffset
PulsePeriod = unwrap.PulsePeriod
ErrorLimitedLookupTable = unwrap.ErrorLimitedLookupTable
LookupTable = unwrap.LookupTable
LookupTableFilename = unwrap.LookupTableFilename


# Custom types


class Analyzer(sciline.Scope[RunType, sc.DataGroup[Any]], sc.DataGroup[Any]):
    """A single wavelength analyzer loaded from an NXcrystal.

    Can be obtained from ``Analyzers``.
    """


class Analyzers(sciline.Scope[RunType, sc.DataGroup[Any]], sc.DataGroup[Any]):
    """All wavelength analyzers loaded from a NXcrystals."""


class DataAtSample(sciline.Scope[RunType, sc.DataArray], sc.DataArray): ...


class DataGroupedByRotation(sciline.Scope[RunType, sc.DataArray], sc.DataArray): ...


class QDetector(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Detector counts with a momentum transfer coordinate."""


EnergyBins = NewType('EnergyBins', sc.Variable)


class EnergyData(sciline.Scope[RunType, sc.DataArray], sc.DataArray): ...


class EnergyQDetector(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Detector counts with a momentum transfer and energy transfer coordinates."""


class IncidentEnergyDetector(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Detector counts with an incident energy coordinate."""


class NormalizedIncidentEnergyDetector(
    sciline.Scope[RunType, sc.DataArray], sc.DataArray
):
    """Normalized detector counts with an incident energy coordinate."""


class SampleAngle(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Rotation angle of the sample, possibly as a function of time.

    For BIFROST, this is angle "a3".
    """


class InstrumentAngle(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Rotation angle of the instrument, possibly as a function of time.

    For BIFROST, this is angle "a4".
    """


NXspeFileName = NewType('NXspeFileName', str)
NXspeFileNames = NewType('NXspeFileNames', list[NXspeFileName])

OutFilename = NewType("OutFilename", str)


class PrimarySpecCoordTransformGraph(sciline.Scope[RunType, dict], dict): ...


class SecondarySpecCoordTransformGraph(sciline.Scope[RunType, dict], dict): ...


InelasticCoordTransformGraph = NewType('InelasticCoordTransformGraph', dict)


class ElasticCoordTransformGraph(sciline.Scope[RunType, dict], dict): ...


class MonitorCoordTransformGraph(sciline.Scope[RunType, dict], dict): ...


SQWBinSizes = NewType('SQWBinSizes', dict[str, int])


UncertaintyBroadcastMode = _UncertaintyBroadcastMode


class ProtonCharge(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Accumulated proton charge for a measurement."""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Utilities for computing real neutron time-of-flight for indirect geometry."""

from collections.abc import Iterable

import sciline
import scippnexus as snx

from ess.reduce import unwrap as reduce_unwrap
from ess.reduce.unwrap.types import DetectorLtotal

from ..types import (
    DataAtSample,
    ErrorLimitedLookupTable,
    LookupTable,
    LookupTableRelativeErrorThreshold,
    MonitorCoordTransformGraph,
    MonitorLtotal,
    MonitorType,
    NeXusDetectorName,
    PulseStrideOffset,
    RawDetector,
    RawMonitor,
    RunType,
    WavelengthDetector,
    WavelengthMonitor,
)


def TofWorkflow(
    *,
    run_types: Iterable[sciline.typing.Key],
    monitor_types: Iterable[sciline.typing.Key],
) -> sciline.Pipeline:
    workflow = reduce_unwrap.GenericUnwrapWorkflow(
        run_types=run_types,
        monitor_types=monitor_types,
    )
    for provider in providers:
        workflow.insert(provider)
    return workflow


def detector_wavelength_data(
    sample_data: DataAtSample[RunType],
    lookup: ErrorLimitedLookupTable[snx.NXdetector],
    pulse_stride_offset: PulseStrideOffset,
) -> WavelengthDetector[RunType]:
    """
    Convert the time-of-arrival data to wavelength data using a lookup table.

    The output data will have a wavelength coordinate.

    This is a wrapper around
    :func:`ess.reduce.unwrap.detector_wavelength_data`
    for indirect geometry spectrometers.
    """
    result = reduce_unwrap.to_wavelength.detector_wavelength_data(
        detector_data=RawDetector[RunType](sample_data),
        lookup=lookup,
        ltotal=DetectorLtotal(sample_data.coords['L1']),
        pulse_stride_offset=pulse_stride_offset,
    )
    # This is the incident wavelength at the sample.
    result.bins.coords['incident_wavelength'] = result.bins.coords.pop('wavelength')
    del result.bins.coords['event_time_zero']
    return result


def monitor_wavelength_data(
    monitor_data: RawMonitor[RunType, MonitorType],
    lookup: ErrorLimitedLookupTable[MonitorType],
    ltotal: MonitorLtotal[RunType, MonitorType],
    pulse_stride_offset: PulseStrideOffset,
) -> WavelengthMonitor[RunType, MonitorType]:
    """
    Convert the time-of-arrival data to wavelength data using a lookup table.

    The output data will have a wavelength coordinate.

    This is a wrapper around
    :func:`ess.reduce.unwrap.monitor_wavelength_data`
    for indirect geometry spectrometers.
    """
    result = reduce_unwrap.to_wavelength.monitor_wavelength_data(
        monitor_data=monitor_data.rename(t='tof'),
        lookup=lookup,
        ltotal=ltotal,
        pulse_stride_offset=pulse_stride_offset,
    )
    result = result.rename(wavelength='incident_wavelength')
    return result


def compute_monitor_ltotal(
    monitor_data: RawMonitor[RunType, MonitorType],
    coord_transform_graph: MonitorCoordTransformGraph,
) -> MonitorLtotal[RunType, MonitorType]:
    """Compute the path length from the source to the monitor."""
    return MonitorLtotal[RunType, MonitorType](
        monitor_data.transform_coords(
            'Ltotal',
            graph=coord_transform_graph,
            keep_intermediate=False,
            keep_aliases=False,
            rename_dims=False,
        ).coords['Ltotal']
    )


def mask_large_uncertainty_in_lut_detector(
    table: LookupTable,
    error_threshold: LookupTableRelativeErrorThreshold,
) -> ErrorLimitedLookupTable[snx.NXdetector]:
    """
    Mask regions in the wavelength lookup table with large uncertainty using NaNs.

    The threshold is looked up under the key ``'detector'``.
    The same threshold is applied to all triplets.

    Parameters
    ----------
    table:
        Lookup table with wavelength as a function of distance and time-of-arrival.
    error_threshold:
        Threshold for the relative standard deviation (coefficient of variation) of the
        projected wavelength above which values are masked.

    See also
    --------
    essreduce.unwrap.mask_large_uncertainty_in_lut:
        The underlying implementation.
    """
    from ess.reduce.unwrap.to_wavelength import (
        mask_large_uncertainty_in_lut_detector,
    )

    return ErrorLimitedLookupTable[snx.NXdetector](
        mask_large_uncertainty_in_lut_detector(
            table=table,
            error_threshold=error_threshold,
            detector_name=NeXusDetectorName('detector'),
        )
    )


providers = (
    compute_monitor_ltotal,
    detector_wavelength_data,
    mask_large_uncertainty_in_lut_detector,
    monitor_wavelength_data,
)
"""Providers for wavelength calculation on indirect geometry spectrometers.

The providers here override the default providers of
:class:`ess.reduce.unwrap.GenericUnwrapWorkflow`
to customize the workflow for indirect geometry spectrometers.
"""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Normalization routines."""

import scipp as sc

from ess.reduce.uncertainty import broadcast_uncertainties
from ess.spectroscopy.types import (
    FrameMonitor3,
    IncidentEnergyDetector,
    NormalizedIncidentEnergyDetector,
    ProtonCharge,
    RunType,
    UncertaintyBroadcastMode,
    WavelengthMonitor,
)


def normalize_by_monitor_and_proton_charge(
    detector: IncidentEnergyDetector[RunType],
    monitor: WavelengthMonitor[RunType, FrameMonitor3],
    proton_charge: ProtonCharge[RunType],
    uncertainty_broadcast_mode: UncertaintyBroadcastMode,
) -> NormalizedIncidentEnergyDetector[RunType]:
    """Normalize detector data by a monitor and proton charge.

    This function divides the detector event weights by the distribution of the monitor
    in incident wavelength and by the proton charge. The former accounts for the
    wavelength-dependent transmission of the beam through the primary spectrometer.
    The latter accounts for the measurement time and source strength.

    The detector is normalized according to

    .. math::

        d_e^\\text{Norm} = d_e \\frac{\\Delta x_i}{m_i \\sum_j m_j} \\frac1{C}

    where :math:`d_e` is a detector event, :math:`m_i` is the monitor bin containing
    the incident wavelength of that detector event, and :math:`\\Delta x_i` is the bin
    width of that bin. Finally, :math:`C` is the proton charge.
    The monitor term is chosen to have unit integral and represent a probability
    density of incident neutrons.

    Parameters
    ----------
    detector:
        Detector events to normalize.
    monitor:
        Monitor histogram.
    proton_charge:
        Accumulated proton charge.
        Scalar or per setting (a3 / a4).
    uncertainty_broadcast_mode:
        Choose how uncertainties of the monitor and proton charge
        are broadcast to the sample data.

    Returns
    -------
    :
        The detector events normalized by the monitor and proton charge.
    """
    # We should make this function public or move the normalization to ESSreduce.
    from ess.reduce.normalization import _mask_detector_for_norm

    if 'time' in monitor.dims:
        # Use a combined monitor distribution for all instrument settings.
        monitor = monitor.sum('time')

    detector = _mask_detector_for_norm(detector=detector, monitor=monitor)

    norm = _monitor_distribution(monitor=monitor)
    # Combine monitor and proton charge so we only operate on events once.
    norm *= proton_charge
    norm = broadcast_uncertainties(
        norm, prototype=detector, mode=uncertainty_broadcast_mode
    )
    return NormalizedIncidentEnergyDetector[RunType](
        detector.bins / sc.lookup(norm, dim=monitor.dim)
    )


def _monitor_distribution(
    monitor: sc.DataArray,
) -> sc.DataArray:
    coord = monitor.coords[monitor.dim]
    delta_w = sc.DataArray(coord[1:] - coord[:-1], masks=monitor.masks)
    return monitor / delta_w / sc.values(monitor).sum()


providers = (normalize_by_monitor_and_proton_charge,)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Live data reduction workflows for BIFROST."""

from collections.abc import Callable
from dataclasses import dataclass
from typing import NewType

import sciline
import scipp as sc

from ess.spectroscopy.types import (
    EnergyQDetector,
    NeXusDetectorName,
    RunType,
)

from .types import ArcEnergy
from .workflow import BifrostWorkflow


@dataclass(frozen=True, kw_only=True, slots=True)
class CutAxis:
    """Axis and bins for cutting 4D Q - delta E data.

    Each axis defines a projection of the :math:`Q` - :math:`\\Delta E` space
    onto a 1D line as well as bin edges on that line.

    Examples
    --------
    Cut along :math:`Q_x`: (see also :meth:`CutAxis.from_q_vector`)

        >>> from ess.bifrost.live import CutAxis
        >>> axis = CutAxis(
        ...     output='Qx',
        ...     fn=lambda sample_table_momentum_transfer: sc.dot(
        ...         sc.vector([1, 0, 0]),
        ...         sample_table_momentum_transfer,
        ...     ),
        ...     bins=sc.linspace(dim='Qx', start=-0.5, stop=0.5, num=100, unit='1/Å'),
        ... )

    Cut along the norm :math:`|Q|`:
    (Note that ``sc.norm`` is wrapped in a lambda to use the
    proper name for the input coordinate, see :attr:`CutAxis.fn`.)

        >>> axis = CutAxis(
        ...     output='|Q|',
        ...     fn=lambda sample_table_momentum_transfer: sc.norm(
        ...         sample_table_momentum_transfer
        ...     ),
        ...     bins=sc.linspace(dim='|Q|', start=-0.9, stop=3.0, num=100, unit='1/Å'),
        ... )

    Cut along :math:`\\Delta E`:

        >>> axis = CutAxis(
        ...     output='E',
        ...     fn=lambda energy_transfer: energy_transfer,
        ...     bins=sc.linspace('E', -0.1, 0.1, 300, unit='meV')
        ... )
    """

    output: str
    """Name of the output coordinate."""
    fn: Callable[[...], sc.Variable]
    """Function to perform the cut.

    Used in :func:`scipp.transform_coords` and so should request input coordinates
    by name.
    """
    bins: sc.Variable
    """Bin edges for the cut."""

    @classmethod
    def from_q_vector(cls, output: str, vec: sc.Variable, bins: sc.Variable):
        """Construct from an arbitrary direction in Q."""
        vec = vec / sc.norm(vec)
        return cls(
            output=output,
            fn=lambda sample_table_momentum_transfer: sc.dot(
                vec, sample_table_momentum_transfer
            ),
            bins=bins,
        )


CutAxis1 = NewType('CutAxis1', CutAxis)
"""Sciline domain type for cut axis 1."""
CutAxis2 = NewType('CutAxis2', CutAxis)
"""Sciline domain type for cut axis 1."""


class CutData(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Data that was cut along CutAxis1 and CutAxis2."""


def arc_energy() -> ArcEnergy:
    """Provide the BIFROST arc energies.

    Returns
    -------
    :
        The final energies for each of the 5 BIFROST analyzer arcs.
    """
    return ArcEnergy(
        sc.array(dims=['arc'], values=[2.7, 3.2, 3.8, 4.4, 5.0], unit='meV')
    )


def cut(
    data: EnergyQDetector[RunType],
    *,
    axis_1: CutAxis1,
    axis_2: CutAxis2,
    arc_energy: ArcEnergy,
) -> CutData[RunType]:
    """Cut data along two axes.

    This function projects the input ``data`` expressed in :math:`Q` and
    :math:`\\Delta E` onto a 2D surface defined by the cut axes.
    This integrates over the other dimensions while preserving the arc dimension
    if present. Then, the projected data is histogrammed according to the axis bins.

    Parameters
    ----------
    data:
        Input data with coordinates "sample_table_momentum_transfer" and
        "energy_transfer".
    axis_1:
        Defines the projection onto and binning in the first axis.
    axis_2:
        Defines the projection onto and binning in the second axis.
    arc_energy:
        The final energies for each analyzer arc.

    Returns
    -------
    :
        ``data`` projected and histogrammed along the cut axes. If the input has
        an arc dimension, it is preserved in the output. Otherwise, all dimensions
        are integrated.
    """
    # Determine which dimensions to preserve
    # Only preserve 'arc' dimension if present; concat everything else
    preserve_dim = 'arc' if 'arc' in data.dims else None

    # Concat over all dimensions except the one we want to preserve
    if preserve_dim is not None:
        dims_to_concat = [dim for dim in data.dims if dim != preserve_dim]
    else:
        dims_to_concat = list(data.dims)

    new_coords = {axis_1.output, axis_2.output}
    projected = data.bins.concat(dims_to_concat).transform_coords(
        new_coords,
        graph={axis_1.output: axis_1.fn, axis_2.output: axis_2.fn},
        keep_inputs=False,
    )
    projected = projected.drop_coords(list(set(projected.coords.keys()) - new_coords))

    # Add arc coordinate if we're working with arc dimension
    # Note that this is for identifying the arc, it should NOT be used
    # as a replacement for a precise final_energy coordinate!
    if preserve_dim == 'arc':
        projected = projected.assign_coords(arc=arc_energy)

    return CutData[RunType](
        projected.hist({axis_2.output: axis_2.bins, axis_1.output: axis_1.bins})
    )


def BifrostQCutWorkflow(detector_names: list[NeXusDetectorName]) -> sciline.Pipeline:
    """Workflow for BIFROST to compute cuts in Q-E-space."""
    workflow = BifrostWorkflow(detector_names)
    workflow.insert(arc_energy)
    workflow.insert(cut)
    return workflow
